unit nrcomm;
// ============================================================
// nrComm Delphi 3 component for talk with Commport device.
// version 0.07
// Full source follow this comment-text.
// Not expelled mistakes and bugs.
// Author: Roman Novgorodov.
// e-mail: noro@chat.ru
//         noro@tomcat.ru
// url:http://www.chat.ru/~noro/
// Freeware price: $0.00
// For more information read nrcomm.hlp
// For example of component use look DEMO.DPR
// Attention:
// If you made application using TnrComm component, or used the
// source code of TnrComm component, you promise to include author
// name "R. Novgorodov" in list of authors of this application project.
//

interface
uses Windows, Messages, SysUtils, Classes, Forms, stdctrls, inifiles;

Type

  TCommEvent=Record      // easy event record type
    Reserved:DWORD;
    len:DWORD;
    str:PChar;
  end;

  TEventList=class      // original list object.
  private
   fCount:longint;
   List:PPointerList;
   function GetEventString(Index:integer):string;
   procedure SetEventString(Index:integer;const S:string);
   function GetEvent(Index:integer):TCommEvent;
   procedure SetEvent(Index:integer; Ev:TCommEvent);
   procedure ClearStr(i:integer);
   procedure ClearEvent(i:integer);
  public
   constructor Create;
   destructor Destroy;   override;
   procedure Add(Ev:Pointer);
   procedure AddNewEvent(Buf:PChar;Len:longint);
   procedure Delete(i:integer);
   procedure Clear;
   property EventString[Index:integer]:string read GetEventString
                                              write SetEventString;
   property Event[Index:integer]:TCommEvent read GetEvent
                                            write SetEvent; default;
   property Count:longint read fCount write fCount;
  end;


  TCommNumbers = 1..8;
  TOnRecived=procedure(Buffer:Pointer;Recived:Cardinal) of object;

  TOnListEvent=procedure(I:longint;List:TEventList)of object;

  TBaudRate = (cbr110, cbr300, cbr600, cbr1200, cbr2400, cbr4800, cbr9600,
    cbr14400, cbr19200, cbr38400, cbr56000, cbr57600, cbr115200, cbr128000, cbr256000);
  TComPort = (cpCOM1,cpCOM2,cpCOM3,cpCOM4);

  TCommThread=class // original easy thread object.
  public
    hDevice,hTerminal:THandle;
    DeviceOL:TOverlapped;     //
    DeviceMask:DWORD;       // Event Mask.
    TotalRecivedSize:dword;        // Recived bytes counter.
    TerminalOnLine:boolean;
    Buffer:PChar;
    Stat:PComStat;            // CommPort Statistic Structure.
    Terminated:boolean;       // Terminated Flag.
    TerminalLocked:boolean;
    EoLnOnLine:boolean;
    hThread:THandle;
    idThread:DWORD;
    DCB:TDCB;                 //
    Paused:boolean;           // Paused flag
    LastChar:char;
    EventList:TEventList;
    EventListOnLine:boolean;
    OnEventList:TOnListEvent;
    OnRecived:TOnRecived;
    constructor Create;   virtual;
    destructor Destroy;   override;
    procedure Suspend;
    procedure Resume;
    procedure Execute;
  end;

  TnrComm = class(TComponent)          // component object .
  private
    hCom:THandle;   // handle of CommPort Device
    FOnActive:TNotifyEvent;
    fTotalSended:dword;
    procedure SetActive(Value:boolean);
    procedure SetTerminalOnLine(Value:boolean);
    function  GetTerminalOnLine:boolean;
    procedure SetTerminal(Value:TMemo);
    procedure SetEventList(Value:TEventList);
    function  GetEventList:TEventList;
  protected
    CommCfg:PCommConfig;
    FMemo:TMemo;
    fBaud:TBaudRate;
    fComPort:TComPort;
    fList:TEventList;
    procedure SetDCB(Value:TDCB);
    function GetDCB:TDCB;
    procedure SetBaudRate(Value:TBaudRate);
    function GetBaudRate:TBaudRate;
    function GetTerminalLocked:boolean;
    procedure SetListOnLine(Val:boolean);
    function GetListOnLine:boolean;
    function GetOnListEvent:TOnListEvent;
    procedure SetOnListEvent(Value:TOnListEvent);
    function GetOnOnRecived:TOnRecived;
    procedure SetOnRecived(V:TOnRecived);
    function GetTotalRecived:dword;
    function GetEolnOnLine:boolean;
    procedure SetEolnOnLine(V:boolean);
  public
    CommThread:TCommThread;
    ActiveComm:boolean;
    OnLine:boolean;
    CommIniSection:string;
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy;                    override;
    procedure SendChar(Ch:Char);
    procedure SendData(Buff:PChar;N:longint);      //Copy
    procedure SendString(S:string);
    procedure ConfigDialog;
    procedure ResetModem;
    procedure SaveProperty(f:TIniFile);      // Ini File store.
    procedure LoadProperty(f:TIniFile);
    property DCB:TDCB read GetDCB write SetDCB;
    property TerminalLocked:boolean read GetTerminalLocked;
    property TotalRecived:dword read GetTotalRecived;
    property TotalSended:dword read fTotalSended;
    property EventList:TEventList read GetEventList write SetEventList;
    property Handle:THandle read hCom;
  published
    property Active:boolean read ActiveComm write SetActive;
    property Terminal:TMemo read fMemo      write SetTerminal;
    property TerminalOnLine:boolean read GetTerminalOnLine
                                    write SetTerminalOnLine;
    property BaudRate:TBaudRate read GetBaudRate write SetBaudRate;
    property ComPort:TComPort read fComPort write fComPort;

    property EoLnOnLine:boolean read GetEolnOnLine write SetEolnOnLine;
    property EventListOnLine:boolean read  GetListOnLine
                                     write SetListOnLine;
    // Events ...
    property OnListEvent:TOnListEvent read  GetOnListEvent
                                      write SetOnListEvent;
    property OnRecived:TOnRecived read GetOnOnRecived
                                  write SetOnRecived;
    property OnActive:TNotifyEvent   read FOnActive write FOnActive;
  end;

  procedure nrDelay(ms:dword); // Easy delay routine ...
  function ThreadProc(Thread:TCommThread):DWORD;   stdcall;
  procedure Register;

implementation

Const
  CommBaudRates: array[TBaudRate] of dword =
    (CBR_110, CBR_300, CBR_600, CBR_1200, CBR_2400, CBR_4800, CBR_9600,
     CBR_14400, CBR_19200, CBR_38400, CBR_56000, CBR_57600, CBR_115200,
     CBR_128000, CBR_256000);


procedure nrDelay(ms:dword);
var tick:dword;
begin
  tick:=GetTickCount;  tick:=tick+ms;//  Maybe this is unnecessary ? ...
  while tick>GetTickCount do Application.ProcessMessages;
end;

//================================================================
//--------------  TEventList class -------------------------------
constructor TEventList.Create;
begin
  List:=Nil;
  fCount:=0;
end;

procedure TEventList.Add(Ev:Pointer);
begin
  ReallocMem(List,(fCount+1)*4);
  List^[fCount]:=Ev;
  Inc(fCount);
end;

procedure TEventList.AddNewEvent(Buf:PChar;Len:longint);
var ce:^TCommEvent;
begin
  New(ce);
  Add(ce);
  GetMem(ce^.str,Len);
  ce^.str:=StrCopy(ce^.str,Buf);
  ce^.len:=Len;
  ce^.Reserved:=0;
end;


procedure TEventList.Delete(i:integer);
begin
  if (I<0)or(I>=fCount) then exit;
  ClearEvent(i);
  if i+1<fCount then Move(List^[i+1],List^[i],(fCount-i)*4);
  Dec(fCount);
end;

procedure TEventList.ClearStr(i:integer);
begin
  if TCommEvent((List^[i])^).str<>Nil
     then FreeMem(TCommEvent((List^[i])^).str);
end;

procedure TEventList.ClearEvent(i:integer);
begin
 ClearStr(i);
 Dispose(List^[i]);
end;

procedure TEventList.SetEventString(Index:integer;const S:string);
begin
  if (Index<0)or(Index>=fCount) then exit;
  ClearStr(index);
  GetMem(TCommEvent((List^[Index])^).str,Length(S)+1);
  TCommEvent((List^[Index])^).str:=StrCopy(TCommEvent((List^[Index])^).str,PChar(S));
  TCommEvent((List^[Index])^).len:=Length(S);
end;

function TEventList.GetEventString(Index:integer):string;
begin
  if (Index>=0)and(Index<fCount) then Result:=String(TCommEvent((List^[Index])^).str)
  else Result:='';
end;

function TEventList.GetEvent(Index:integer):TCommEvent;
begin
 if (Index>=0)and(Index<fCount) then Result:=TCommEvent((List^[Index])^)
 else exit;
end;

procedure TEventList.SetEvent(Index:integer; Ev:TCommEvent);
begin
  if (Index<0)or(Index>=fCount) then exit;
  ClearEvent(index);
  TCommEvent((List^[Index])^):=Ev;
end;

procedure TEventList.Clear;
var i:integer;
begin
  for i:=0 to fCount-1 do Delete(0);
end;

destructor TEventList.Destroy;
begin
  Clear;
  FreeMem(List);
end;

//================================================================
//===============    TnrComm    ==================================
constructor TnrComm.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  New(CommCfg);
  ActiveComm:=False;
  CommThread:=TCommThread.Create;
  CommIniSection:='Com Port Options';
  EventList:=TEventList.Create;
end;

procedure TnrComm.SetActive(Value:boolean);
var CommTimeouts:TCommTimeouts;
begin
  CommThread.Suspend;
  if Value then begin
    CommThread.TotalRecivedSize:=0;
    fTotalSended:=0;
    if hCom<>0 then CloseHandle(hCom);
    // !!! Try to Open ComPortDevice !!!
    hCom:=CreateFile(PChar('COM'+IntToStr(Integer(ComPort)+1)), // device name
          GENERIC_READ + GENERIC_WRITE,// acces mode
          0,                      // not shared
          Nil,                    // not security
          OPEN_EXISTING,          // already exist
          FILE_FLAG_OVERLAPPED,   // anisochronous
          0);   //
    if hCom=INVALID_HANDLE_VALUE then Exit;

    // can be not necessary ...
    FillChar(CommTimeouts,SizeOf(CommTimeouts),0);
    CommTimeouts.ReadIntervalTimeout:=MAXDWORD;
    SetCommTimeouts(hCom,CommTimeouts);

    GetCommState(hCom,CommCfg^.DCB); // get current config ...
    CommCfg^.DCB.BaudRate:=CommBaudRates[fBaud];// change config ...
    SetCommState(hCom,CommCfg^.DCB);   // set changed config ...

    CommThread.hDevice:=hCom;
    CommThread.Resume;   // Go on ...
    if Assigned(OnActive) then OnActive(Self);
  end else begin
    if hCom<>0 then CloseHandle(hCom);
    hCom:=0;
  end;
  ActiveComm:=Value;
end;

function TnrComm.GetTerminalLocked:boolean;
begin  Result:=CommThread.TerminalLocked;                      end;
procedure TnrComm.SetTerminalOnLine(Value:boolean);
begin  CommThread.TerminalOnLine:=Value;                       end;
function TnrComm.GetTerminalOnLine:boolean;
begin  Result:=CommThread.TerminalOnLine;                      end;

procedure TnrComm.SetListOnLine(Val:boolean);
begin CommThread.EventListOnLine:=Val                          end;
function TnrComm.GetListOnLine:boolean;
begin  Result:=CommThread.EventListOnLine;                     end;
procedure TnrComm.SetEventList(Value:TEventList);
begin  CommThread.EventList:=Value;                            end;
function  TnrComm.GetEventList:TEventList;
begin  Result:=CommThread.EventList;                           end;

procedure TnrComm.SetTerminal(Value:TMemo);
begin fMemo:=Value; CommThread.hTerminal:=Value.Handle;        end;
procedure TnrComm.SetDCB(Value:TDCB);
begin  CommCfg^.DCB:=Value;  SetCommState(hCom,CommCfg^.DCB);  end;
function TnrComm.GetDCB:TDCB;
begin   GetCommState(hCom,CommCfg^.DCB); Result:=CommCfg^.DCB; end;

function TnrComm.GetOnListEvent:TOnListEvent;
begin Result:=CommThread.OnEventList;                          end;

procedure TnrComm.SetOnListEvent(Value:TOnListEvent);
begin CommThread.OnEventList:=Value;                           end;

function TnrComm.GetEolnOnLine:boolean;
begin  Result:=CommThread.EoLnOnLine;                          end;

procedure TnrComm.SetEolnOnLine(V:boolean);
begin CommThread.EoLnOnLine:=V;                                end;


procedure TnrComm.SendChar(Ch:Char);
begin
  if Active then TransmitCommChar(hCom,Ch);
  Inc(fTotalSended);
end;

procedure TnrComm.SendData(Buff:PChar;N:longint);
var Nn:dword;
begin
  if Active then WriteFile(hCom,Buff^,N,Nn,Nil);
  fTotalSended:=fTotalSended+Nn;
end;

procedure TnrComm.SendString(S:string);
begin  SendData(PChar(S+#13),Length(S)+1); end;

procedure TnrComm.SetBaudRate(Value:TBaudRate);
begin
  GetCommState(hCom,CommCfg^.DCB);
  CommCfg^.DCB.BaudRate:=CommBaudRates[Value];
  SetCommState(hCom,CommCfg^.DCB);
  fBaud:=Value;
end;

function TnrComm.GetBaudRate:TBaudRate;
var i:integer;
begin
  Result:=cbr56000;
  GetCommState(hCom,CommCfg^.DCB);
  for i:=0 to 14 do
      if CommCfg^.DCB.BaudRate=CommBaudRates[TBaudRate(i)]
      then begin
        Result:=TBaudRate(i);
        break;
      end;
end;

procedure TnrComm.ConfigDialog;
var n:dword;
begin  // call buildin config dialog ...
  n:=SizeOf(CommCfg^);
  GetCommConfig(hCom,CommCfg^,N);
  CommConfigDialog(PChar('COM'+IntToStr(Integer(ComPort)+1)),Application.Handle,CommCfg^);
  SetCommConfig(hCom,CommCfg^,N);
end;

procedure TnrComm.ResetModem;
begin
  EscapeCommFunction(hCom,CLRDTR);
  EscapeCommFunction(hCom,CLRRTS);
  nrDelay(1000);
  Active:=False;
  Active:=True;
end;

procedure TnrComm.SaveProperty(f:TIniFile);
begin with f do begin
  WriteInteger(CommIniSection,'ComPort',Integer(ComPort));
  WriteInteger(CommIniSection,'BaudRate',Integer(BaudRate));
end;end;

procedure TnrComm.LoadProperty(f:TIniFile);
begin with f do begin
  ComPort:=TComPort(ReadInteger(CommIniSection,'ComPort',Integer(cpCOM1)));
  BaudRate:=TBaudRate(ReadInteger(CommIniSection,'BaudRate',Integer(BaudRate)));
end;end;

function TnrComm.GetTotalRecived:dword;
begin  Result:=CommThread.TotalRecivedSize; end;

function TnrComm.GetOnOnRecived:TOnRecived;
begin Result:=CommThread.OnRecived;         end;
procedure TnrComm.SetOnRecived(V:TOnRecived);
begin CommThread.OnRecived:=V;              end;

destructor TnrComm.Destroy;
begin
  CommThread.Destroy;
  Dispose(CommCfg);
  if hCom<>0 then CloseHandle(hCom);
  EventList.Destroy;
  inherited Destroy;
end;

//================================================================
//=============== CommThread =====================================
constructor TCommThread.Create;
begin
  New(Stat);
  Terminated:=False;
  TerminalLocked:=False;
  LastChar:=#0;
  Paused:=True;
  DeviceOL.Offset:=0;
  DeviceOL.OffsetHigh:=0;
  DeviceOL.hEvent:=CreateEvent(Nil,False,False,Nil);
  { mask of event ... you may try any other combinations ...}
  DeviceMask:=EV_RXCHAR;//or EV_BREAK or EV_CTS or EV_DSR or EV_ERR or EV_RING
                        //or EV_RLSD or EV_RXFLAG or EV_TXEMPTY
  EoLnOnLine:=True;
  hThread:=CreateThread(Nil,0,@ThreadProc,Pointer(Self),CREATE_SUSPENDED,idThread);
end;

destructor TCommThread.Destroy;
begin
  Terminated:=not Paused;
  while Terminated do;  // correct terminate thread ...
  CloseHandle(hThread); // Close Thread Handle.
  CloseHandle(DeviceOL.hEvent);  // close event object.
  Dispose(Stat);
end;

procedure TCommThread.Suspend; // Paused thread execution ...
begin if not Paused then SuspendThread(hThread);Paused:=True;end;
procedure TCommThread.Resume;  // Paused thread execution ...
begin if Paused then ResumeThread(hThread);Paused:=False; end;

procedure TCommThread.Execute;
var r,Err:dword;
    Mask:dword;
    i,ii:dword;
begin
  while not Terminated do begin
    SetCommMask(hDevice,DeviceMask);
    WaitCommEvent(hDevice,Mask,@DeviceOL);
    repeat
      r:=WaitForSingleObject(DeviceOL.hEvent,0);
      if Terminated then Break;
      Application.ProcessMessages;
    until r<>WAIT_TIMEOUT;
    if r=WAIT_OBJECT_0 then case Mask of
     EV_RXCHAR:      // !!! event of data recive
       if ClearCommError(hDevice,Err,Stat) then begin
         TotalRecivedSize:=TotalRecivedSize+Stat^.cbInQue;
         //TotalSendSize:=TotalSendSize+Stat^.cbOutQue;
         Buffer:=StrAlloc(Stat^.cbInQue);
         ReadFile(hDevice,Buffer^,Stat^.cbInQue,i,Nil); //read data

         // Terminal works ...
         if TerminalOnLine then begin // if Terminal exsists ...
           TerminalLocked:=True;
           for i:=0 to Stat^.cbInQue-1 do begin // EoLn control ...
             if (EoLnOnLine)and(Buffer[i]=#$0a)and(LastChar=#$0d)
                then begin
                  LastChar:=Buffer[i];
                  continue;
             end;
             SendMessage(hTerminal,WM_CHAR,Byte(Buffer[i]),0);
             LastChar:=Buffer[i];
           end;
           TerminalLocked:=False;
         end; // if TerminalOnLine

         // EventList works ...
         if EventListOnLine and Assigned(EventList)then
           for i:=0 to Stat^.cbInQue-1 do for ii:=0 to EventList.fCount-1 do begin
              r:=TCommEvent((EventList.List^[ii])^).Reserved;
              if Buffer[i]=TCommEvent((EventList.List^[ii])^).str[r]
                 then Inc(r) else r:=0;
              TCommEvent((EventList.List^[ii])^).Reserved:=r;
              if r<TCommEvent((EventList.List^[ii])^).len then continue;
              TCommEvent((EventList.List^[ii])^).Reserved:=0;
              if Assigned(OnEventList) then OnEventList(ii,EventList);
         end;

         // On recived events ...
         if Assigned(OnRecived) then OnRecived(Buffer,Stat^.cbInQue);

         StrDispose(Buffer);
       end; // if ...
    end; // case ...
   end; // while ...
   Terminated:=False;
end;

//=================================================================
//=================================================================
function ThreadProc(Thread:TCommThread):DWORD;
begin  Thread.Execute;  Result:=0 end;
//=================================================================
//=================================================================
procedure Register;
begin
 RegisterComponents('NoRo', [TnrComm]);
end;
end.
